use crate::models::Group;
use pyo3::prelude::*;
use pyo3::exceptions::PyRuntimeError;
use postgres::{Client, NoTls};
use crate::models::User;

fn pg_err(e: postgres::Error) -> PyErr {
    PyRuntimeError::new_err(e.to_string())
}

/// Create a group
#[pyfunction]
pub fn create_group(db_url: &str, name: &str) -> PyResult<()> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    client.execute(
        "INSERT INTO groups (name, member_count) VALUES ($1, $2)", // Use $2 for member_count
        &[&name, &0]  // Pass 0 explicitly as an i32
    ).map_err(pg_err)?;
    Ok(())
}

/// Retrieve a group by ID.
#[pyfunction]
pub fn get_group(db_url: &str, group_id: i32) -> PyResult<Option<Group>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let row_opt = client.query_opt(
        "SELECT id, name, member_count FROM groups WHERE id = $1",
        &[&group_id]
    ).map_err(pg_err)?;
    
    if let Some(row) = row_opt {
        let group = Group {
            id: row.get(0),
            name: row.get(1),
            member_count: row.get(2),
        };
        Ok(Some(group))
    } else {
        Ok(None)
    }
}

/// Retrieve all groups.
#[pyfunction]
pub fn get_all_groups(db_url: &str) -> PyResult<Vec<Group>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let rows = client.query("SELECT id, name, member_count FROM groups", &[])
        .map_err(pg_err)?;
    
    let groups = rows.into_iter().map(|row| Group {
        id: row.get(0),
        name: row.get(1),
        member_count: row.get(2),
    }).collect();
    
    Ok(groups)
}

/// Add a user to a group.
#[pyfunction]
pub fn add_user_to_group(db_url: &str, group_id: i32, user_id: i32) -> PyResult<()> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    client.execute(
        "INSERT INTO group_members (group_id, user_id) VALUES ($1, $2)",
        &[&group_id, &user_id]
    ).map_err(pg_err)?;
    client.execute(
        "UPDATE groups SET member_count = member_count + 1 WHERE id = $1",
        &[&group_id]
    ).map_err(pg_err)?;
    Ok(())
}

/// Get all members of a group.
#[pyfunction]
pub fn get_group_members(db_url: &str, group_id: i32) -> PyResult<(Vec<User>, i32)> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;

    // Fetch group members
    let rows = client.query(
        "SELECT u.id, u.name, u.email, u.username, u.studentid, u.startyear, u.endyear, u.profilepicture, u.firstlogin
         FROM users u
         JOIN group_members gm ON u.id = gm.user_id
         WHERE gm.group_id = $1",
        &[&group_id]
    ).map_err(pg_err)?;

    let users = rows.into_iter().map(|row| User {
        id: row.get(0),
        name: row.get(1),
        email: row.get(2),
        username: row.get(3),
        studentid: row.get(4),
        startyear: row.get(5),
        endyear: row.get(6),
        profilepicture: row.get(7),
        firstlogin: row.get(8),
    }).collect();

    // Fetch member count
    let count: i32 = client.query_one(
        "SELECT member_count FROM groups WHERE id = $1",
        &[&group_id]
    ).map_err(pg_err)?.get(0);

    Ok((users, count))
}


/// Group Delete Function
#[pyfunction]
pub fn delete_group(db_url: &str, group_id: i32) -> PyResult<bool> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let deleted = client.execute(
        "DELETE FROM groups WHERE group_id = $1",
        &[&group_id]
    ).map_err(pg_err)?;
    Ok(deleted > 0)
}

#[pyfunction]
pub fn search_groups(db_url: &str, query: &str) -> PyResult<Vec<Group>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let rows = client.query(
        "SELECT id, name, member_count FROM groups WHERE name ILIKE $1",
        &[&format!("%{}%", query)] 
    ).map_err(pg_err)?;

    let groups: Vec<Group> = rows.iter()
        .map(|row| Group {
            id: Some(row.get(0)),
            name: row.get(1),
            member_count: row.get(2),
        })
        .collect();

    Ok(groups)
}
