use crate::models::User;
use pyo3::prelude::*;
use pyo3::exceptions::PyRuntimeError;
use postgres::{Client, NoTls};
use crate::auth::hash_password;
use crate::auth::verify_password;
use postgres::types::ToSql;

fn pg_err(e: postgres::Error) -> PyErr {
    PyRuntimeError::new_err(e.to_string())
}

#[pyfunction]
pub fn create_user(
    db_url: &str, name: &str, email: &str, password: &str, username: &str, studentid: i32, startyear: i32, endyear: i32, profilepicture: Option<&str>, firstlogin: bool) -> PyResult<()> {
    let password_hash = hash_password(password)?;
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let result = client.execute(
        "INSERT INTO users (name, email, password_hash, username, studentid, startyear, endyear, profilepicture, firstlogin)
         VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9)",
        &[&name, &email, &password_hash, &username, &studentid, &startyear, &endyear, &profilepicture, &firstlogin],
    );

    match result {
        Ok(_) => {
            Ok(())
        }
        Err(e) => {
            // Look for a Postgres UNIQUE violation (error code 23505)
            if let Some(code) = e.code() {
                if code.code() == "23505" {
                    return Err(PyRuntimeError::new_err("Student ID already exists."));
                }
            }
            // Otherwise, return the generic error
            Err(pg_err(e))
        }
    }
}

/// Retrieve a user by ID. Returns None if the user is not found.
#[pyfunction]
pub fn get_user(db_url: &str, user_id: i32) -> PyResult<Option<User>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let row_opt = client.query_opt(
        "SELECT id, name, email, username, studentid, startyear, endyear, profilepicture, firstlogin FROM users WHERE id = $1",
        &[&user_id]
    ).map_err(pg_err)?;
    
    if let Some(row) = row_opt {
        let user = User {
            id: row.get(0),
            name: row.get(1),
            email: row.get(2),
            username: row.get(3),
            studentid: row.get(4),
            startyear: row.get(5),
            endyear: row.get(6),
            profilepicture: row.get(7),
            firstlogin: row.get(8),
            
        };
        Ok(Some(user))
    } else {
        Ok(None)
    }
}

/// Retrieve all users from the database.
#[pyfunction]
pub fn get_all_users(db_url: &str) -> PyResult<Vec<User>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let rows = client.query("SELECT id, name, email, username, studentid, startyear, endyear, profilepicture, firstlogin FROM users", &[])
        .map_err(pg_err)?;
    
    let mut users = Vec::new();
    for row in rows {
        users.push(User {
            id: row.get(0),
            name: row.get(1),
            email: row.get(2),
            username: row.get(3),
            studentid: row.get(4),
            startyear: row.get(5),
            endyear: row.get(6),
            profilepicture: row.get(7),
            firstlogin: row.get(8),

        });
    }
    Ok(users)
}

// Only include optional parameters in the `signature` attribute
#[pyfunction(signature = (
    db_url,
    user_id,
    new_name = None,
    new_email = None,
    new_password = None,
    new_username = None,
    new_studentid = None,
    new_startyear = None,
    new_endyear = None,
    new_profilepicture = None
))]
pub fn update_user(
    db_url: &str,  // Required, do NOT include in signature
    user_id: i32,  // Required, do NOT include in signature
    new_name: Option<&str>,
    new_email: Option<&str>,
    new_password: Option<&str>,
    new_username: Option<&str>,
    new_studentid: Option<i32>,
    new_startyear: Option<i32>,
    new_endyear: Option<i32>,
    new_profilepicture: Option<&str>,
) -> PyResult<()> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let mut query = String::from("UPDATE users SET ");
    let mut updates = Vec::new();
    
    // Owned parameters to avoid temporary borrow issues
       
    let mut owned_params: Vec<Box<dyn ToSql + Sync>> = Vec::new();

    macro_rules! push_field {
        ($field:expr, $value:expr) => {
            if let Some(v) = $value {
                updates.push(format!("{} = ${}", $field, owned_params.len() + 1));
                owned_params.push(Box::new(v));
            }
        };
    }

    push_field!("name", new_name);
    push_field!("email", new_email);
    
    if let Some(password) = new_password {
        let password_hash = hash_password(password)?;
        updates.push(format!("password_hash = ${}", owned_params.len() + 1));
        owned_params.push(Box::new(password_hash));
    }
    
    push_field!("username", new_username);
    push_field!("studentid", new_studentid);
    push_field!("startyear", new_startyear);
    push_field!("endyear", new_endyear);
    push_field!("profilepicture", new_profilepicture);

    if updates.is_empty() {
        return Err(PyRuntimeError::new_err("No fields to update."));
    }

    query.push_str(&updates.join(", "));
    query.push_str(&format!(" WHERE id = ${}", owned_params.len() + 1));
    owned_params.push(Box::new(user_id));

    let params: Vec<&(dyn ToSql + Sync)> = owned_params.iter().map(|b| &**b).collect();
    let updated = client.execute(query.as_str(), &params).map_err(pg_err)?;
    
    if updated == 0 {
        Err(PyRuntimeError::new_err("User not found"))
    } else {
        Ok(())
    }
}

/// Delete a user by ID. Returns true if a user was deleted.
#[pyfunction]
pub fn delete_user(db_url: &str, user_id: i32) -> PyResult<bool> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let deleted = client.execute(
        "DELETE FROM users WHERE id = $1",
        &[&user_id]
    ).map_err(pg_err)?;
    Ok(deleted > 0)
}

///Verify credentials
///Retrieve a stored password hash for the email and compare.
#[pyfunction]
pub fn verify_user(db_url: &str, email: &str, password: & str) -> PyResult<bool> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let row_opt = client.query_opt(
        "SELECT password_hash FROM users WHERE email = $1",
        &[&email]
        ).map_err(pg_err)?;
        if let Some(row) = row_opt {
            let stored_hash: String = row.get(0);
            verify_password(&stored_hash, password)
            } else {
                Ok(false)
        }
}

#[pyfunction]
pub fn update_user_profile_picture(db_url: &str, user_id: i32, path: &str) -> PyResult<()> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let updated = client.execute(
        "UPDATE users SET profilepicture = $1 WHERE id = $2",
        &[&path, &user_id],
    ).map_err(pg_err)?;
    if updated == 0 {
        Err(PyRuntimeError::new_err("User not found"))
    } else {
        Ok(())
    }
}

#[pyfunction]
pub fn update_first_login(db_url: &str, user_id: i32, first_login: bool) -> PyResult<()> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let updated = client.execute(
        "UPDATE users SET firstlogin = $1 WHERE id = $2",
        &[&first_login, &user_id],
    ).map_err(pg_err)?;
    if updated == 0 {
        Err(PyRuntimeError::new_err("User not found"))
    } else {
        Ok(())
    }
}
